/* -*- mode: c++; c-basic-offset: 3; indent-tabs-mode: nil; -*- */

// System Header Files   
#include <sys/statvfs.h>
#ifdef __sun
#include <sys/time.h>
#include <sys/lwp.h>
#endif /* __sun */

#include <errno.h>

#include <algorithm>
#include <iostream>
#include <iterator>
#include <limits>
#include <set>
#include <sstream>
#include <stdexcept>
#include <vector>

#include "general/AtExit.hh"
#include "general/Thread.hh"

// GenericAPI
#include "genericAPI/MountPointStatus.hh"
#include "genericAPI/Logging.hh"
#include "genericAPI/StatVFS.hh"

// Local Header Files
#include "diskcacheAPI.hh"
#include "mountpt.hh"
#include "MountPointCache.hh"
#include "MountPointScanner.hh"
#include "dirhash.hh"   
#include "dirinfo.hh"   
#include "cachefile.hh"   
#include "filename.hh"   
#include "Query.hh"
#include "QueryResultsFilenames.hh"
#include "QueryResultsIntervals.hh"
#include "QueryResultsMountPointStats.hh"
#include "Symbols.hh"

static bool
initialize( );

namespace diskCache
{
   namespace Private
   {
      namespace MountPt_CC
      {
      }
   }
}

using namespace diskCache;
using namespace diskCache::Private::MountPt_CC;
using namespace General;
using namespace std;
using GenericAPI::MountPointStatus;
using GenericAPI::StatVFS;

typedef struct statvfs stat_vfs_type;
   
#define DEBUG_DISKCACHE_MOUNT_PT_TIMES 0
/// \cond IGNORE
#if 0
#include "general/IOLock.hh"

#define VERBOSE_DEBUGGING 1
#define	AT_BASE( a ) \
{ \
   int oldstate; \
   pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &oldstate ); \
   MutexLock    lock( General::IOLock::GetKey( std::cerr ) ); \
   pthread_setcancelstate( oldstate, &oldstate ); \
\
   std::cerr << a << __FILE__ << " " << __LINE__ << std::endl << std::flush; \
}

#else
#define	AT_BASE( a )
#endif
#define	AT( a )	AT_BASE( a << std::endl << "\t" )
#define	HERE( ) AT_BASE( "" )
/// \endcond

static void
doStatVFS( const std::string& Path, stat_vfs_type& Buffer,
           const int Line, const char* const File )
{
   //------------------------------------------------------------
   // Setup as if StatVFS is NOT going to be called
   //------------------------------------------------------------
   int err = -1;
   bool enabled = false;
 
   diskCache::Symbols::DISKCACHE_ENABLE_STATVFS::Get( enabled);

   if ( enabled )
   {
      err = StatVFS( Path.c_str( ), Buffer );
   }
   else
   {
      errno = EINTR;
   }

   if ( err == -1 )
   {
      std::ostringstream msg;
            
      msg << "StatVFS: " << Path
          << " err: " << err
          << " errno: " << errno
          << " line: " << Line
          << " file: " << File
         ;
      throw std::runtime_error( msg.str( ) );
   }
}

class print_element
{
public:
   print_element( std::string& Stream, const char Spacer, bool First = true )
      : m_stream( Stream ),
        m_spacer( Spacer ),
        m_first( First )
   {
   }

   void operator()( const std::string& Data )
   {
      if ( ! m_first )
      {
         m_stream += m_spacer;
      }
      else
      {
         m_first = false;
      }
      m_stream += Data;
   }

private:
   mutable std::string& m_stream;
   const char           m_spacer;
   mutable bool         m_first;
};

//-----------------------------------------------------------------------
/// \brief Predicate class to validate list of mount points.
//-----------------------------------------------------------------------
class validator
{
public:
   typedef std::string element_type;

   //--------------------------------------------------------------------
   /// \brief Constructor
   //--------------------------------------------------------------------
   validator( const std::vector< element_type >& Source );

   void Errors( std::string& Message ) const;
   //--------------------------------------------------------------------
   /// \brief Unary condition operator
   //--------------------------------------------------------------------
   bool operator()( const element_type& MountPoint );

   //--------------------------------------------------------------------
   /// \brief Binary condition operator
   //--------------------------------------------------------------------
   bool operator()( const element_type& Current,
                    const element_type& Previous );

private:
   struct state_type {
      std::set< element_type >          m_child_directories;
      std::set< element_type >          m_duplicate_directories;
      std::map< element_type, INT_4U >  m_dir_count;
   };
   //--------------------------------------------------------------------
   /// \brief Check parental status
   ///
   /// \param[in] Current
   ///     String to check if it is a subdirectory of Previous
   /// \param[in] Previous
   ///     String used in parental check
   ///
   /// \return
   ///     The value of true is returned if Current is a subdirectory
   ///     of Previous.
   //--------------------------------------------------------------------
   static bool is_parent( const element_type& Current,
                          const element_type& Previous );

   const std::list< element_type >              m_original_directories;
   std::list< element_type >                    m_parent_directories;
   General::SharedPtr< state_type >             m_state;
};

namespace diskCache
{
   namespace Private
   {
      //-----------------------------------------------------------------
      /// \brief Private resouces for mountpt.cc
      ///
      /// This namespace is intended for resources that are needed only
      /// within mountpt.cc.
      //-----------------------------------------------------------------
      namespace MountPt_CC
      {
         //--------------------------------------------------------------
         /// \brief Auxiliary mount point information
         ///
         /// This class does the book keeping of auxiliary information
         /// which needs to be maintained about each mount point.
         /// Most of this information is to enhance performance.
         //--------------------------------------------------------------
         class AuxInfo {
         public:
            //-----------------------------------------------------------
            /// \brief Default constructor
            ///
            /// This is the default constructor. Key members are
            /// explicately initialized while the rest use their
            /// default constructors.
            //-----------------------------------------------------------
            AuxInfo( )
               : m_initialized( false )
            {
            }


            //-----------------------------------------------------------
            /// \brief Invalidate the data because of conflict
            //-----------------------------------------------------------
            void
            Invalidate( )
            {
               m_initialized = false;
            }

            //-----------------------------------------------------------
            /// \brief Return the file system modification status
            //-----------------------------------------------------------
            bool
            Modified( ) const
            {
               stat_vfs_type buffer;

               try
               {
                  doStatVFS( m_path, buffer, __LINE__, __FILE__ );

                  if ( m_initialized )
                  {
#define CHECK( a ) ( buffer.a == m_statvfs_buffer.a )
                     if ( CHECK( f_bfree )
                          && CHECK( f_bavail )
                          && CHECK( f_files )
                          && CHECK( f_ffree )
                          && CHECK( f_favail ) )
                     {
                        return false;
                     }
#undef CHECK
                  }
                  //-----------------------------------------------------
                  // Copy the stat info for future scans
                  //-----------------------------------------------------
                  m_statvfs_buffer = buffer;
                  m_initialized = true;
               }
               catch( const std::runtime_error& Error )
               {
                  switch( errno )
                  {
                  case EINTR:
                  case ENOSYS:
                  case EOVERFLOW:
                     //--------------------------------------------------
                     // Quietly absorb the exception and report back
                     // that the directory has changed since the
                     // file system does not support the call.
                     //--------------------------------------------------
                     return true;
                  case EACCES:
                  case EFAULT:
                  case ENAMETOOLONG:
                  case ENOENT:
                     //--------------------------------------------------
                     // Quietly absorb the exception and report back
                     // that nothing has changed as this directory
                     // is not accessable
                     //--------------------------------------------------
                     return false;
                     break;
                  default:
                     //--------------------------------------------------
                     // Rethrow the exception for someone else to
                     // handle
                     //--------------------------------------------------
                     throw;
                  }
               }
               return true;
            }

            //-----------------------------------------------------------
            /// \brief Syncronize with the current state of the
            ///        file system.
            ///
            /// \return
            ///     Nothing
            //-----------------------------------------------------------
            void
            Sync( )
            {
               try
               {
                  doStatVFS( m_path, m_statvfs_buffer, __LINE__, __FILE__ );
                  m_initialized = true;
               }
               catch( const std::runtime_error& Error )
               {
                  switch( errno )
                  {
                     case EACCES:
                     case EFAULT:
                     case ENAMETOOLONG:
                     case ENOENT:
                        //-----------------------------------------------
                        // Quietly absorb the exception and report back
                        // proper initialization since this directory
                        // is not accessable
                        //-----------------------------------------------
                        m_initialized = true;
                        break;
                  }
               }
            }

            //-----------------------------------------------------------
            /// \brief Get file associated with the auxiliary information.
            ///
            /// \return
            ///     The path associated with this auxiliary information.
            //-----------------------------------------------------------
            inline const std::string&
            Path( ) const
            {
               return m_path;
            }

            //-----------------------------------------------------------
            /// \brief Set path to be associated with auxiliary information.
            ///
            /// \param Filename
            ///     Path to associate with auxiliary information.
            ///
            /// \return
            ///     Nothing
            //-----------------------------------------------------------
            inline void
            Path( const std::string& Filename )
            {
               m_path = Filename;
            }

         private:
            //-----------------------------------------------------------
            /// \brief Typedef to normalize system specific implementation.
            ///
            /// To help minimize the customization of the code,
            /// the system specific type used to detect modifications
            //  to the file system is aliased to stat_vfs_type.
            //-----------------------------------------------------------
            typedef struct statvfs stat_vfs_type;

            //-----------------------------------------------------------
            /// \brief Path associated with the file system information.
            ///
            /// The directory or file used in the system calls to
            /// obtain the file system information.
            //-----------------------------------------------------------
            std::string         m_path;

            //-----------------------------------------------------------
            /// \brief Buffer containing cached information
            ///
            /// This storage space caches the state of the file system
            /// at the time of the last sync. It is used in determining
            /// the "dirty" state of the file system.
            //-----------------------------------------------------------
            mutable stat_vfs_type       m_statvfs_buffer;

            //-----------------------------------------------------------
            /// \brief Initialized state.
            ///
            /// Keep track of the initialized state of m_statvfs_buffer.
            //-----------------------------------------------------------
            mutable bool                m_initialized;
         }; // class - AuxInfo

         //--------------------------------------------------------------
         /// \brief Collection of additional file system information.
         //--------------------------------------------------------------
         class AuxInfoSet
            : protected General::unordered_map< std::string,
                                                General::SharedPtr< AuxInfo > >
         {
         public:
            typedef General::unordered_map< std::string,
                                                General::SharedPtr< AuxInfo > >
            container_type;

            typedef container_type::key_type key_type;
            typedef container_type::mapped_type mapped_type;
            typedef container_type::value_type value_type;

            //-----------------------------------------------------------
            /// \brief Removes directories from being managed.
            //-----------------------------------------------------------
            inline void
            Remove( const std::vector< std::string >& Dirs )
            {
               for ( std::vector< std::string >::const_iterator
                        cur = Dirs.begin( ),
                        last = Dirs.end( );
                     cur != last;
                     ++cur )
               {
                  erase( find( *cur ) );
               }
            }

            inline void
            erase( container_type::iterator Pos )
            {
               MutexLock        l( m_baton );

               container_type::erase( Pos );
            }

            inline size_t
            erase( const key_type& Key )
            {
               MutexLock        l( m_baton );

               return container_type::erase( Key );
            }

            inline container_type::mapped_type
            operator[]( const key_type& Key )
            {
               MutexLock        l( m_baton );

               mapped_type& retval( container_type::operator[]( Key ) );
               if ( retval.get( ) == NULL )
               {
                  retval.reset( new mapped_type::element_type );

                  retval->Path( Key );
               }
               return retval;
            }

         protected:
            static MutexLock::lock_type       m_baton;
         }; // class - AuxInfoSet


         static const CHAR FrameNameSeparator( FileName::getFrameNameSeparator() );
   
   
         template < class T >
         class compare_range
         {
         public:
            int
            operator( )( const std::pair< T, T >& T1,
                         const std::pair< T, T >& T2 ) const
            {
               return T1.first < T2.first;
            }
         };

         template< class T >
         class TimeRange
         {
         public:
            typedef std::vector< pair< T, T > > range_type;

            TimeRange( const std::vector< T >& Points )
            {
               for ( typename std::vector< T >::const_iterator
                        cur = Points.begin( ),
                        last = Points.end( );
                     cur != last;
                     ++cur )
               {
                  m_ranges.push_back( pair< T, T >( *cur++, *cur ) );
               }
               //--------------------------------------------------------------
               // Sort the data based on the start time of the interval
               //--------------------------------------------------------------
               sort( m_ranges.begin( ), m_ranges.end( ),
                     compare_range< T >( ) );
               //--------------------------------------------------------------
               // :TODO: Merge continuous segments togeter
               //--------------------------------------------------------------
               for ( typename range_type::iterator
                        cur = m_ranges.begin( );
                     cur != m_ranges.end( );
                     )
               {
                  for ( typename range_type::iterator
                           cur2 = cur+1;
                        cur2 != m_ranges.end( );
                        )
                  {
                     if ( cur2->second <= cur->second )
                     {
                        //-----------------------------------------------------
                        // Complete subset so just get rid of it
                        //-----------------------------------------------------
                        cur2 = m_ranges.erase( cur2 );
                     }
                     else if ( ( cur2->first >= cur->first ) &&
                               ( cur2->first <= cur->second ) )
                     {
                        //-----------------------------------------------------
                        // Merge these two together as there is overlap
                        //-----------------------------------------------------
                        cur->second = cur2->second;
                        cur2 = m_ranges.erase( cur2 );
                     }
                     else
                     {
                        break;
                     }
                  }
                  //-----------------------------------------------------------
                  // Advance to the next time range
                  //-----------------------------------------------------------
                  ++cur;
               }
            }

            std::ostream&
            operator << ( std::ostream& Stream ) const
            {
               for ( typename range_type::const_iterator
                        cur = m_ranges.begin( ),
                        last = m_ranges.end( );
                     cur != last;
                     ++cur )
               {
                  Stream << " " << cur->first << " " << cur->second;
               }
               return Stream;
            }
         private:
            range_type        m_ranges;
         };

         //: Predicate used to check if string contains any forward slashes
         //: (it's full or partial path).
         class IsPath : public unary_function< string, bool >
         {
         public:
      
            IsPath( std::string& errors );
   
            bool operator()( const string& path ) const;
   
   
         private:

            //: A reference to the accumulative error message
            std::string& mErrors;
         };   
   
   
         //: Constructor
         //
         //!param: const string& errors - An accumulative error message.
         //
         IsPath::IsPath( std::string& errors )
            : mErrors( errors )
         {}
   
   
         //: Compare operator
         //
         //!param: const string& path - A string to validate.
         //
         //!return: bool - True if "path" is full or partial path, false otherwise.
         //
         bool IsPath::operator()( const string& path ) const
         {
            static const CHAR slash( '/' );  
   
            if( path.find( slash ) != string::npos )
            {
               // FIX for PR 2603: opening brace needs preceding space
               mErrors += " {Ignored provided pattern \'";
               mErrors += path;
               mErrors += "\' (Slashes are detected, only explicit directory names are supported.)} ";
   
               return true;
            }
   
            return false;
         }
   
   
         //: Predicate used to check if directory has a specific
         //: parent directory (the same as DirectoryHash::HasParent + error logging)
         class LogParent : public DirectoryHash::HasParent
         {
         public:
      
            LogParent( const string& parent, std::string& errors );
   
            bool operator()( const string& dir ) const;
   
   
         private:

            //: Accumulative error message
            std::string& mErrors;
         };
   
   
         //: Constructor
         //
         //!param: const string& parent - Parent directory name.
         //!param: string& errors - A reference to the accumulative error.   
         //
         LogParent::LogParent( const string& parent, std::string& errors )
            : HasParent( parent ),
              mErrors( errors )
         {}
   
   
         //: Compare operator
         //
         //!param: const string& path - Full path of the directory to examine.
         //
         //!return: bool - True if directory has specified parent directory,
         //+        false otherwise.
         //
         bool LogParent::operator()( const string& path ) const
         {
            if( HasParent::operator()( path ) )
            {
               // Log error
               mErrors += path;
               // FIX for PR 2603: opening brace needs preceding space
               mErrors += " {Directory is ignored as a subdirectory of existing MOUNT_PT entry \'";
               mErrors += mParent;
               mErrors += "\'} ";
   
               return true;
            }
   
            return false;
         }      
   
   
         //: Predicate used to check if a pair of ifo and type 
         //: satisfy a searched criteria: "IFO" part contains searched character,
         //: "Type" is an exact match of searched "type".
         class HasIfoCharAndType : public unary_function< string, bool >
         {
         public:
      
            HasIfoCharAndType( const CHAR& ifo, const string& type );
   
            bool operator()( const string& ifo_type ) const;
   
   
         private:

            //: Searched ifo character for frame data
            CHAR mIfo; 
      
            //: A requested type for frame data
            string mType;
         };

   
         //: Constructor
         //
         //!param: const CHAR& ifo - Searched IFO character for frame data.
         //!param: const string& type - Searched Type for frame data.
         //
         HasIfoCharAndType::HasIfoCharAndType( const CHAR& ifo, const string& type )
            : mIfo( ifo ), mType( type )
         {
            // Include FrameNameSeparator to guarantee exact type match:
            mType += FrameNameSeparator;         
         }
   
   
         //: Compare operator
         //
         //!param: const string& ifo_type - "IFO-Type-" formated string..
         //
         //!return: bool - True if key matches searched ifo and type, false otherwise.
         //
         bool HasIfoCharAndType::operator()( const string& ifo_type ) const
         {
            static const CHAR* const method_name( "HasIfoCharAndType::operator(): " );

   
            // Locate first name delimiter: between "IFO" and "Type"
            string::size_type dash_pos( ifo_type.find( FrameNameSeparator ) );
   
            // Sanity check: key should always be properly formatted, if not -
            // something is really wrong!!!
            if( dash_pos == string::npos )
            {
               string msg( method_name );
               msg += "Could not extract Type information from hash map key \'";
               msg += ifo_type;
               msg += "\'";
   
               throw runtime_error( msg.c_str() );
            }
   
   
            // Check if ifo part of the key contains mIfo character
            string::size_type ifo_pos = ifo_type.find( mIfo );
   
            if( ifo_pos == string::npos || 
                ifo_pos >= dash_pos )
            {
               return false;
            }

   
            // Compare "Type" part of the key:
            return ( ifo_type.compare( ++dash_pos, mType.size(), mType ) == 0 );
         }

         template< class T >
         std::ostream&
         operator<<( std::ostream& Stream,  const TimeRange< T >& Source )
         {
            return Source.operator<<( Stream );
         }

         //**************************************************************
         // Local data
         //**************************************************************
         MutexLock::lock_type
         AuxInfoSet::m_baton = MutexLock::Initialize( );
         //--------------------------------------------------------------
         /// \brief Storage for additional file system information.
         //--------------------------------------------------------------
         static AuxInfoSet     aux_fs_info;
      } // namespace - MountPt_CC
   } // namespace - Private
} // namespace - diskCache
   
   
namespace diskCache
{
   //--------------------------------------------------------------------
   // Static data members definition   
   //--------------------------------------------------------------------

   bool MountPointHash::initialized =
      initialize( );

   MutexLock::lock_type MountPointHash::m_scanned_mount_points_baton
      = MutexLock::Initialize( );

   INT_4U MountPointHash::m_scanned_mount_points = 0;

   MutexLock::lock_type MountPointHash::m_last_update_baton
      = MutexLock::Initialize( );

   General::GPSTime MountPointHash::m_last_update;

   const INT_4U MountPointHash::SCANNED_MOUNT_POINTS_RESET
     = std::numeric_limits< INT_4U >::max( );

   MountPointHash::lock_type::lock_type MountPointHash::mLock
     = MountPointHash::lock_type::Sync::INITIALIZER;

   MountPointHash::lock_type::lock_type MountPointHash::mLockFileExtList
     = MountPointHash::lock_type::Sync::INITIALIZER;

   bool MountPointHash::mDataConflictsAcrossAllMntPts = false;

   std::list< std::string > MountPointHash::mExcludedDirList;
   
   std::list< std::string > MountPointHash::mFileExtList;
   
   MountPointHash::MountPtContainer MountPointHash::mMountPtHash;

   MountPointHash::excluded_dirs_container_type MountPointHash::p_excluded_dirs_container;
   MountPointHash::excluded_dirs_rw_type::lock_type
   MountPointHash::p_excluded_dirs_container_baton = ReadWriteLock::Initialize( );
   
   MountPointHash::mount_point_container_type MountPointHash::p_mount_point_container;
   MountPointHash::mount_point_ro_type::lock_type
   MountPointHash::p_mount_point_container_baton = ReadWriteLock::Initialize( );
   


   //===================================================================
   //====================================================================
   MountPointHash::cache_container_base_type::
   cache_container_base_type( )
      : m_baton( lock_type::Sync::INITIALIZER )
   {
   }

   MountPointHash::cache_container_base_type::
   ~cache_container_base_type( )
   {
      clear( );
   }

   inline void MountPointHash::cache_container_base_type::
   Find( const Query& SearchParams,
         QueryResults& Results )
   {
      lock_type l( m_baton,
                   lock_type::READ,
                   __FILE__, __LINE__ );

      if ( SearchParams.Criteria( ) & Query::CRITERIA_BIT_FILENAME_HANDLER )
      {
         //--------------------------------------------------------------
         // Look for a specifc handler
         //--------------------------------------------------------------
         cache_container_base_type::const_iterator pos_handler
            = find( SearchParams.FNHandler( ) );

         if ( ( pos_handler == end( ) )
              || ( pos_handler->second.get( )
                   == (MountPointCache*)NULL ) )
         {
            //----------------------------------------------------------
            // Since it cannot be found, continue with the next
            // entry
            //----------------------------------------------------------
            return;
         }
         pos_handler->second->Find( SearchParams, Results );
      }
      else
      {
         Query       sp( SearchParams );
         //--------------------------------------------------------------
         // Loop over all file handlers
         //--------------------------------------------------------------
         for ( cache_container_base_type::const_iterator
                  cur_handler = begin( ),
                  last_handler = end( );
               cur_handler != last_handler;
               ++cur_handler )
         {
            if ( cur_handler->second )
            {
               sp.FNHandler( cur_handler->first );
               cur_handler->second->Find( SearchParams, Results );
            }
         }

      }
   }

   void MountPointHash::cache_container_base_type::
   Register( dir_info_type DirInfo )
   {
#if ENABLE_MOUNT_POINT_CACHE
      if ( DirInfo
           && DirInfo->HasData( ) )
      {
         lock_type l( m_baton,
                      lock_type::WRITE,
                      __FILE__, __LINE__ );

         //--------------------------------------------------------------
         // di is used to get rid of the const of DirInfo.
         //--------------------------------------------------------------

         General::SharedPtr< DirectoryInfo >       di( DirInfo );
         DirectoryInfo::index_variable_rw_type     i( di->Index( ) );
         
         for ( DirectoryInfo::index_type::iterator
                  cur = i.Var( ).begin( ),
                  last = i.Var( ).end( );
               cur != last;
               ++cur )
         {
            iterator pos = find( cur->first );

            if ( pos == end( ) )
            {
               insert( value_type( cur->first, cur->first->CreateMountPointCache( ) ) );
               pos = find( cur->first );
            }
            //--------------------------------------------------------------
            // Process request
            //   pos->second == MountPointCache Information
            //   cur->second == DirectoryInfo compressed information
            //--------------------------------------------------------------
            pos->second->Register( cur->second );
         }
      }
#endif /* ENABLE_MOUNT_POINT_CACHE */
   }

   void MountPointHash::cache_container_base_type::
   Replace( dir_info_type DirInfoOld,
            dir_info_type DirInfoNew )
   {
#if ENABLE_MOUNT_POINT_CACHE
#if WORKING
      if ( DirInfoOld.get( )
           && DirInfoNew.get( )
           && DirInfoNew->HasData( ) )
      {
         lock_type l( m_baton,
                      lock_type::WRITE,
                      __FILE__, __LINE__ );

         //--------------------------------------------------------------
         //--------------------------------------------------------------
         General::SharedPtr< DirectoryInfo >       din( DirInfoNew );
         DirectoryInfo::index_variable_rw_type     in( din->Index( ) );
         General::SharedPtr< DirectoryInfo >       dio( DirInfoOld );
         DirectoryInfo::index_variable_rw_type     io( dio->Index( ) );
         
         
         //--------------------------------------------------------------
         // Remove old entry
         //--------------------------------------------------------------
         if ( DirInfoOld.get( ) != NULL )
         {
            for ( DirectoryInfo::index_type::iterator
                     cur = io.Var( ).begin( ),
                     last = io.Var( ).end( );
                  cur != last;
                  ++cur )
            {
               iterator pos = find( cur->first );

               if ( pos != end( ) )
               {
                  pos->second->Unregister( cur->second );
                  erase( pos );
               }
            }
         }
         //-----------------------------------------------------------------
         // Add new entry
         //-----------------------------------------------------------------
         if ( DirInfoNew->HasData( ) )
         {
            for ( DirectoryInfo::index_type::iterator
                     cur = in.Var( ).begin( ),
                     last = in.Var( ).end( );
                  cur != last;
                  ++cur )
            {
               iterator pos = find( cur->first );

               if ( pos == end( ) )
               {
                  insert( value_type( cur->first, cur->first->CreateMountPointCache( ) ) );
                  pos = find( cur->first );
               }
               //-----------------------------------------------------------
               // Process request
               //   pos->second == MountPointCache Information
               //   cur->second == DirectoryInfo compressed information
               //-----------------------------------------------------------
               pos->second->Register( cur->second );
            }
         }
      }
      else
      {
         Unregister( DirInfoOld );
         Register( DirInfoNew );
      }
#else /* WORKING */
      Unregister( DirInfoOld );
      Register( DirInfoNew );
#endif /* WORKING */
#endif /* ENABLE_MOUNT_POINT_CACHE */
   }

   void MountPointHash::cache_container_base_type::
   Unregister( dir_info_type DirInfo )
   {
#if ENABLE_MOUNT_POINT_CACHE
      if ( DirInfo.get( ) == NULL )
      {
         //--------------------------------------------------------------
         // Do not go off into the weeds simply because there is no
         // data associated with the DirInfo.
         //--------------------------------------------------------------
         return;
      }

      lock_type l( m_baton,
                   lock_type::WRITE,
                   __FILE__, __LINE__ );

      typedef std::pair< key_type, mount_pt_cache_type > remove_element_type;
      typedef std::list< remove_element_type > remove_container_type;
      remove_container_type                    remove;

      {
         //--------------------------------------------------------------
         // Go through and remove references to the directory information.
         //--------------------------------------------------------------      
         General::SharedPtr< DirectoryInfo >       di( DirInfo );
         DirectoryInfo::index_variable_rw_type     i( di->Index( ) );
         
         for ( DirectoryInfo::index_type::iterator
                  cur = i.Var( ).begin( ),
                  last = i.Var( ).end( );
               cur != last;
               ++cur )
         {
            iterator pos = find( cur->first );

            if ( pos != end( ) )
            {
               pos->second->Unregister( cur->second );
               //--------------------------------------------------------
               // Looks like this memory can be reclaimed
               //--------------------------------------------------------
               if ( ( pos->second.get( ) == NULL )
                    || ( pos->second->empty( ) ) )
               {
                  remove.push_back( remove_element_type( cur->first, pos->second ) );
               }
            }
         }
         //--------------------------------------------------------------
         // Check to see if resource can be reclaimed.
         //--------------------------------------------------------------
         if ( remove.empty( ) == false )
         {
            //-----------------------------------------------------------
            // Reclaim resources
            //-----------------------------------------------------------
            General::ReadWriteLock l( mLock,
                                      General::ReadWriteLock::WRITE,
                                      __FILE__, __LINE__ );
            for ( remove_container_type::const_iterator
                     cur = remove.begin( ),
                     last = remove.end( );
                  cur != last;
                  ++cur )
            {
               iterator pos = find( cur->first );
               
               //--------------------------------------------------------
               // Verify that the node should still be cleaned.
               //--------------------------------------------------------
               if ( ( pos != end( ) )
                    && ( pos->second )
                    && ( pos->second == cur->second )
                    || ( pos->second->empty( ) ) )
               {
                  erase( pos );
               }
            }
         }
      }
#endif /* ENABLE_MOUNT_POINT_CACHE */
   }

   //====================================================================
   // Class - MountPointHash
   //====================================================================

   inline MountPointHash::excluded_dirs_rw_type MountPointHash::
   excluded_dirs( )
   {
      return( excluded_dirs_rw_type( p_excluded_dirs_container_baton,
                                     RWLOCK_TIMEOUT,
                                     p_excluded_dirs_container,
                                     __FILE__, __LINE__ ) )
         ;
   }

   inline MountPointHash::mount_point_rw_type MountPointHash::
   mount_points( )
   {
      return( mount_point_rw_type( p_mount_point_container_baton,
                                   RWLOCK_TIMEOUT,
                                   p_mount_point_container,
                                   __FILE__, __LINE__ ) )
         ;
   }

   void MountPointHash::
   Register( dir_info_type DirInfo )
   {
#if ENABLE_MOUNT_POINT_CACHE
      const std::string key( DirInfo->GetMountPointPath( ).Var( ) );

      MountPtContainer::mapped_type cache_info_mgr;
      
      {
         bool found = false;
         {
            //--------------------------------------------------
            // Try to find the information with only a read
            // lock to help speed up the process.
            //--------------------------------------------------
            ReadWriteLock lock( mLock, ReadWriteLock::READ,
                                __FILE__, __LINE__ );   
   
            //--------------------------------------------------
            // Lookup the information.
            //--------------------------------------------------
            MountPtContainer::iterator i
               = mMountPtHash.find( key );
            found = ( i != mMountPtHash.end( ) );
            if ( found )
            {
               cache_info_mgr = i->second;
            }
         }
         //-----------------------------------------------------
         // Check to see if the mount point was found
         //-----------------------------------------------------
         if ( found == false )
         {
            //--------------------------------------------------
            // Since the object is not currently in the list,
            // need a write lock so as to allow the container
            // to be expanded for the new data.
            //--------------------------------------------------
            ReadWriteLock lock( mLock, ReadWriteLock::WRITE,
                                __FILE__, __LINE__ );   
   
            //--------------------------------------------------
            // Create space for the new object
            //--------------------------------------------------
            cache_info_mgr = mMountPtHash[ key ];
         }
      }

      if ( cache_info_mgr.get( ) == NULL )
      {
         //------------------------------------------------------
         // Lock
         //------------------------------------------------------
         ReadWriteLock lock( mLock, ReadWriteLock::WRITE,
                             __FILE__, __LINE__ );

         //------------------------------------------------------
         // Double check
         //------------------------------------------------------
         cache_info_mgr = mMountPtHash[ key ];
         if ( cache_info_mgr.get( ) == NULL )
         {
            //---------------------------------------------------
            // Create an entry
            //---------------------------------------------------
            cache_info_mgr =
               new MountPtContainer::mapped_type::element_type;
            
            mMountPtHash[ key ] = cache_info_mgr;
         }
   
      }
      cache_info_mgr->Register( DirInfo );
#else   /* ENABLE_MOUNT_POINT_CACHE */
      mMountPtHash[ DirInfo->GetPath( ).Var( ) ] = DirInfo;
#endif /* ENABLE_MOUNT_POINT_CACHE */
      MountPointHash::LastUpdateSet( );
   }

   void MountPointHash::
   Replace( dir_info_type DirInfoOld,
            dir_info_type DirInfoNew )
   {
      //-----------------------------------------------------------------
      // Check for sameness
      //-----------------------------------------------------------------
      if ( DirInfoOld == DirInfoNew )
      {
         //--------------------------------------------------------------
         // Nothing to do since they are the same pointer
         //--------------------------------------------------------------
         return;
      }
#if WORKING
      if ( ( DirInfoOld.get( ) == NULL )
           && ( DirInfoNew.get( ) ) == NULL )
      {
         //--------------------------------------------------------------
         // There is nothing to do here
         // Leave before core dumping.
         //--------------------------------------------------------------
         return;
      }

      const std::string mount_point( ( DirInfoOld.get( ) )
                                     ? DirInfoOld->GetMountPointPath( ).Var( )
                                     : DirInfoNew->GetMountPointPath( ).Var( ) );
      MountPtContainer::mapped_type cache_info_mgr;
      bool found = false;

      {
         ReadWriteLock lock( mLock, ReadWriteLock::READ,
                             __FILE__, __LINE__ );   
   
         MountPtContainer::iterator i = mMountPtHash.find( mount_point );
         found = ( i != mMountPtHash.end( ) );
         if ( found )
         {
            cache_info_mgr = i->second;
         }
         else if ( ( DirInfoNew.get( ) != NULL )
                   && ( DirInfoNew->HasData( ) ) )
         {
            lock.Modify( ReadWriteLock::WRITE,
                         __FILE__, __LINE__ );

            //--------------------------------------------------
            // Create space for the new object
            //--------------------------------------------------
            cache_info_mgr = mMountPtHash[ mount_point ];
         }
      }
      if ( cache_info_mgr.get( ) != NULL )
      {
         cache_info_mgr->Replace( DirInfoOld, DirInfoNew );
         MountPointHash::LastUpdateSet( );
      }
#else  /* WORKING */
      //-------------------------------------------------------------------------
      // If there was existing data, deallocate outside of the mutex lock
      //-------------------------------------------------------------------------
      Unregister( DirInfoOld );
      //-------------------------------------------------------------------------
      // If the new entry has some information, then need to register so it is
      // searchable.
      //-------------------------------------------------------------------------
      Register( DirInfoNew );
#endif /* WORKING */
   }

   void MountPointHash::
   Unregister( dir_info_type DirInfo )
   {
#if ENABLE_MOUNT_POINT_CACHE
      const std::string key( DirInfo->GetMountPointPath( ).Var( ) );

      MountPtContainer::mapped_type cache_info_mgr;
      
      {
         ReadWriteLock lock( mLock, ReadWriteLock::READ,
                             __FILE__, __LINE__ );   
   
         MountPtContainer::iterator i = mMountPtHash.find( key );
         if ( i == mMountPtHash.end( ) )
         {
            return;
         }
         cache_info_mgr = i->second;
      }

      if ( cache_info_mgr.get( ) == NULL )
      {
         return;
      }
      cache_info_mgr->Unregister( DirInfo );
#else   /* ENABLE_MOUNT_POINT_CACHE */
#error Do not know how to unregister data from the mount point hash.
#endif /* ENABLE_MOUNT_POINT_CACHE */
      MountPointHash::LastUpdateSet( );
   }

   void MountPointHash::
   Find( const Query& SearchParams,
         QueryResults& Results )
   {
      const Query::bit_flag_type criteria = SearchParams.Criteria( );

      if ( criteria & Query::CRITERIA_BIT_MOUNT_POINT )
      {
         MountPtContainer::mapped_type data;

         {
            lock_type	l( mLock,
                           lock_type::READ,
                           __FILE__, __LINE__ );

            //----------------------------------------------------------
            // Searching for a specific mount point
            //----------------------------------------------------------
            MountPtContainer::const_iterator
               pos_mount_point( mMountPtHash.find( SearchParams.MountPoint( ) ) );

         
            if ( pos_mount_point == mMountPtHash.end( ) )
            {
               //-------------------------------------------------------
               // Since it cannot be found, do nothing
               //-------------------------------------------------------
               return;
            }
            data = pos_mount_point->second;
         }

         data->Find( SearchParams, Results );
      }
      else
      {
         //--------------------------------------------------------------
         // Step through all MOUNT_PT entries in the proper order
         //--------------------------------------------------------------
         Query  sp( SearchParams );

         const mount_point_container_type
            mpts( MountPoints( ).Var( ) );

         MountPtContainer::mapped_type data;

         for( ordered_mount_point_type::const_iterator
                 cur_mount = mpts.begin( ),
                 last_mount = mpts.end( ); 
              cur_mount != last_mount;
              ++cur_mount )
         {
            {
               lock_type	l( mLock,
                                   lock_type::READ,
                                   __FILE__, __LINE__ );

               // Does mount point have any frame data registered?
               MountPtContainer::const_iterator
                  pos_mount_point( mMountPtHash.find( *cur_mount ) );

               if ( pos_mount_point == mMountPtHash.end( ) )
               {
                  //----------------------------------------------------
                  // Since it cannot be found, continue with the next
                  // entry
                  //----------------------------------------------------
                  continue;
               }
               data = pos_mount_point->second;
            }

            //-----------------------------------------------------------
            // Restrict searching to this mount point
            //-----------------------------------------------------------
            sp.MountPoint( *cur_mount );

            data->Find( SearchParams, Results );
         }
      }
   } // method - Find

   //--------------------------------------------------------------------
   /// This method initializes ordered list of MOUNT_PT directories as
   /// they appear in the API resource file if called at API startup time;
   /// otherwise, it checks if any MOUNT_PT directory has been pruned,
   /// and removes that directory from global hash.   
   //--------------------------------------------------------------------
   const std::string MountPointHash::
   updateMountPtList( const CHAR* const dir_list,
                      const bool enable_global_check )   
   {
      static const CHAR* const
         method_name( "MountPointHash::updateMountPtList(): " );

      // FIX for PR 2603: opening brace needs preceding space
      static const CHAR* const empty_lists( " {} {} " );

      static std::string previous;
   
      if( dir_list == 0 || *dir_list == '\0' )
      {
         string msg( method_name );
         msg += "Empty directory list is specified.";
   
         throw runtime_error( msg.c_str() );
      }
   
      if ( previous.compare( dir_list ) == 0 )
      {
         return empty_lists;
      }

      //-----------------------------------------------------------------
      /// \todo Protect modification of this variable with a lock
      //-----------------------------------------------------------------
         mDataConflictsAcrossAllMntPts = enable_global_check;

      // These vectors will be used only if MOUNT_PT list got updated
      vector< string > new_list, old_list_ordered;

      // Removed subdirectories with frame data
      list< string > all_removed_subdirs;
   
      // Accumulative error message:
      LogMessage errors;   
   
      // Check if any directories got removed from MOUNT_PT list,
      // remove them recursively from global hash:
      vector< string > removed;
   
      insert_iterator< vector< string > >
         remove_iter( removed, removed.begin() );   

      //-----------------------------------------------------------------
      // Check for differences in the mount points
      //-----------------------------------------------------------------
      {
         istringstream list_stream( dir_list );
   
         //--------------------------------------------------------------
         // Parse the input stream creating a list of mount points
         //--------------------------------------------------------------
         copy( istream_iterator< string >( list_stream ),
               istream_iterator< string >( ), 
               back_inserter( new_list ) );  
         //--------------------------------------------------------------
         // Validate new entries: entries in conflict will be
         // removed by this method and captured in "errors"   
         //--------------------------------------------------------------
         errors( ) << validateMountPtList( new_list );
         //--------------------------------------------------------------
         // Get old mount point entries
         //--------------------------------------------------------------
         old_list_ordered = MountPoints( ).Var( );      // read-only
         sort( old_list_ordered.begin( ),
               old_list_ordered.end( ) );   
         //--------------------------------------------------------------
         // Sort the new list
         //--------------------------------------------------------------
         vector< string > new_list_ordered( new_list );
         sort( new_list_ordered.begin( ),
               new_list_ordered.end( ) );   
         //--------------------------------------------------------------
         // Create list of deleted mount points
         //--------------------------------------------------------------
         set_difference( old_list_ordered.begin( ), old_list_ordered.end( ),
                         new_list_ordered.begin( ), new_list_ordered.end( ),
                         remove_iter );      
         
         //--------------------------------------------------------------
         // Check for differences
         // * Size differences
         // * Anything removed
         // * Change in order
         //--------------------------------------------------------------
         if ( ( new_list_ordered.size( ) == old_list_ordered.size( ) )
              && ( removed.empty( ) )
              && ( new_list == MountPoints( ).Var( ) ) )
         {
            //-----------------------------------------------------------
            // No differences so return to the caller
            //-----------------------------------------------------------
            return empty_lists;
         }
      }
      // Introduce this scope for WRITE lock   
      {
         //--------------------------------------------------------------
         // Need to suspend the scanning of mount points.
         //--------------------------------------------------------------
         std::ostringstream     msg;

         msg << "Setting up to modify mount point list"
            ;
         GenericAPI::queueLogEntry( msg.str( ),
                                    GenericAPI::LogHTML::MT_DEBUG,
                                    0,
                                    method_name,
                                    "SCAN_MOUNTPT" );
         MountPointScanner::SyncRequest sync_request;
         MountPointScanner::scanner_sync_rw_type
            scanner_sync_lock( MountPointScanner::Sync( ) );

         mount_point_rw_type    mpts( mount_points( ) );

         MountPointStatus::Set( new_list );
         
         for( vector< string >::iterator
                 cur = removed.begin(),
                 last = removed.end();
              cur != last;
              ++cur )
         {
            // Remove mount point from frame hash, 
            // then remove directory recursively


            list< string > just_removed( removeEntry( *cur, errors ) );
   
            all_removed_subdirs.splice( all_removed_subdirs.end(), just_removed );
         }
   

         // Reset the mMountPtList to be the same.
         // If new directories have been added to the MOUNT_PT list, 
         // Tcl layer is responsible for invoking "getDirEnt_t" thread
         // per each new directory in the list. 
         // C++ layer only removes directories, and resets the ordered list of 
         // MOUNT_PT directories for further look ups.
         mpts.Var( ) = new_list;
   

         previous = dir_list;

         // Return properly formatted Tcl string
         return formatOutput( all_removed_subdirs, errors.Message( ).str( ) );
      }
   
      previous = dir_list;

      if( errors.Message( ).str( ).empty() == false )   
      {
         // Return properly formatted Tcl string
         return formatOutput( all_removed_subdirs, errors.Message( ).str( ) );
      }

   
      return empty_lists;
   }

   
   //--------------------------------------------------------------------
   /// This method updates a list of known directory names that should
   /// not be scanned for any data (don't even call "lstat" for these).
   /// It also scans existing directory hash if such directories already
   /// exist, and removes them from the hash (recursively).   
   //--------------------------------------------------------------------
   const std::string MountPointHash::
   updateExcludedDirList( const CHAR* const dir_list )   
   {
      static const CHAR* const
         method_name( "MountPointHash::updateExcludeDirList: " );

      // FIX for PR 2603: opening brace needs preceding space
      static const CHAR* const empty_lists( " {} {} " );
   
      // Accumulative error message
      string errors;

      static std::string previous;

      if ( previous.compare( dir_list ) == 0 )
      {
         return empty_lists;
      }

      //-----------------------------------------------------------------
      // Need to suspend the scanning of mount points.
      //-----------------------------------------------------------------
      std::ostringstream     msg;

      msg << "Setting up to modify exclude dir list list"
         ;
      GenericAPI::queueLogEntry( msg.str( ),
                                 GenericAPI::LogHTML::MT_DEBUG,
                                 0,
                                 method_name,
                                 "SCAN_MOUNTPT" );
      MountPointScanner::SyncRequest sync_request;
      MountPointScanner::scanner_sync_rw_type
         scanner_sync_lock( MountPointScanner::Sync( ) );

      {
         excluded_dirs_rw_type    edirs( excluded_dirs( ) );
         
         if( dir_list == 0 || *dir_list == '\0' )
         {
            edirs.Var( ).clear( );

            return empty_lists;
         }

   
         {
            edirs.Var( ).clear( );
      
      
            istringstream list_stream( dir_list );
      
      
            // Reset the list
            copy( istream_iterator< string >( list_stream ),
                  istream_iterator< string >(), 
                  back_inserter( edirs.Var( ) ) );

   
            edirs.Var( ).remove_if( IsPath( errors ) );      
         }
      }
      // Check existing directory hash if such directories already exist 
      LogMessage     el;

      list< string > removed_dirs( DirectoryHash::Instance( ).
                                   excludeDirectories( ExcludedDirs( ).Var( ), el ) );
      errors += el.Message( ).str( );

      ostringstream header_stream;
      header_stream << createExcludedMsgHeader( ExcludedDirs( ).Var( ) );

   
      string header( header_stream.str() );
   
   
      previous = std::string( dir_list );
      // Return properly formatted Tcl string
      return formatOutput( removed_dirs, errors, header );
   }
   

   //--------------------------------------------------------------------
   /// This method updates a list of known file extensions 
   //--------------------------------------------------------------------
   void MountPointHash::
   updateFileExtList( const CHAR* const ext_list )
   {
      if( ext_list == 0 || *ext_list == '\0' )
      {
         string msg( "MountPointHash::updateFileExtList(): " );
         msg += "Empty file extension list is specified.";
  
         throw runtime_error( msg.c_str() );
      }

      // Scope for lock
      {
         ReadWriteLock lock( mLockFileExtList, ReadWriteLock::WRITE,
                             __FILE__, __LINE__ );

         // Clear out existing list
         mFileExtList.clear();

         // Insert the raw string into a stream to easily tokenize on whitespace
         stringstream ss(ext_list); 

         string buffer;

         // Tokenize the string on whitespace
         while (ss >> buffer)
         {
            mFileExtList.push_back(buffer);
         }

      }

      return;
   }
   

   //--------------------------------------------------------------------
   /// This method inserts a new record for the directory into the
   /// global hash map if the hash doesn't contain an entry for it.
   /// It invokes resursive scan for the already existing directory
   /// in the hash map.
   ///
   /// \sa getDirEnt
   //--------------------------------------------------------------------
   const std::string MountPointHash::
   addEntry( const std::string& dir )   
   {
      static const char* caller = "diskcacheAPI::MountPonitHash::addEntry";

      ScannedMountPoints( 1 );
      //----------------------------------------------------------------
      // Do some setup if running in a thread
      //--------------------------------------------------------------==
      Thread::Self::CancellationType( Thread::CANCEL_EXCEPTION );
      // Accumulative error message
      LogMessage                errors;
      string                    error_str;
      list< string >            data;

      General::GPSTime          start;
      std::ostringstream        msg;
   
      msg << " dir: " << dir;
      start.Now( );
      AT( "DEBUG: addEntry - start: " << dir );
      msg << " IsDirManaged: " << MountPointStatus::IsDirManaged( dir );

      try
      {
      if ( MountPointStatus::IsDirManaged( dir ) )
      {
         AT( "DEBUG: MountPointStatus::Status( " << dir << " ): " << MountPointStatus::Status( dir ) );
         if ( MountPointStatus::Status( dir )
              == MountPointStatus::STATE_ONLINE )
         {
            //-----------------------------------------------------------
            // Everything seems to be good
            //-----------------------------------------------------------
            AuxInfoSet::mapped_type ai = aux_fs_info[ dir ];
            //------------------------------------------------------------
            // Check if the scan really needs to be done
            //------------------------------------------------------------
            if ( ai->Modified( ) )
            {
               msg << " Scan: Deep";

               //---------------------------------------------------------
               // Insert new entry into the global DirectoryHash, or
               // rescan directory in already existing hash record.
               //---------------------------------------------------------
               data = DirectoryHash::Instance().scanDirectory( dir,
                                                               dir,
                                                               ExcludedDirs( ).Var( ),
                                                               errors );
               if ( errors.Message( ).str( ).empty( ) == false )
               {
                  ai->Invalidate( );
               }
            }
            else
            {
               msg << " Scan: Skip";
            }
         }
         else
         {
            //-----------------------------------------------------------
            // Go through and flag data as unavailable since the device
            //   is offline.
            //-----------------------------------------------------------
            std::list< std::string >       exclude;

            {
               exclude = ExcludedDirs( ).Var( );
            }

            AT( "DEBUG: mtime: " << DirectoryHash::Instance().GetMountPointMTime( dir,
                                                                                  exclude ) );
            if ( DirectoryHash::Instance().GetMountPointMTime( dir, exclude )
                 > 0 )
            {
               data = DirectoryHash::Instance().scanDirectory( dir,
                                                               dir,
                                                               exclude,
                                                               errors );
            }
         }  // if ( MountPointStatus::Status()
      } // if ( MountPointStatus::IsManaged()
      else
      {
         // FIX for PR 2603: opening brace needs preceding space
         errors( ) << " {Specified directory \'"
                   <<  dir
                   << "\' is not in MOUNT_PT list, call updateMountPtList() first.}"
                   << " ERROR: " << error_str;
         errors.MessageType( LogMessage::LM_NOTE );
      } // if ( MountPointStatus::IsManaged()
   
      General::GPSTime          end;
      end.Now( );
      // Return properly formatted Tcl string
      AT( "DEBUG: addEntry - end: " << dir );
      msg << " Elapsed Time: " << ( end - start ) << " seconds";
#if DEBUG_DISKCACHE_MOUNT_PT_TIMES
#ifdef __sun
      struct lwpinfo    lwp;

      if ( _lwp_info( &lwp ) == 0 )
      {
         msg
            << "( user: "
            << ( lwp.lwp_utime.tv_sec
                 + ( double(lwp.lwp_utime.tv_nsec) * 1.0e-9 ) )
            << "s system: "
            << ( lwp.lwp_stime.tv_sec
                 + ( double(lwp.lwp_stime.tv_nsec) * 1.0e-9 ) )
            << "s )";
      }
#endif /* __sun */
#endif /* DEBUG_DISKCACHE_MOUNT_PT_TIMES */
      }
      catch( const std::exception& except )
      {
         errors( ) << "{exception: " << except.what( ) << " }";
      }
      error_str = formatOutput( data, errors.Message( ).str( ) );
      if ( errors.MessageType( ) != LogMessage::LM_DEBUG )
      {
         //--------------------------------------------------------------
         // Write the non empty message to the log stream
         //--------------------------------------------------------------
         GenericAPI::queueLogEntry( errors.Message( ).str( ),
                                    (GenericAPI::LogHTML::message_type)( errors.MessageType( ) ),
                                    0,
                                    caller,
                                    "SCAN_MOUNTPT" );
      }
      //-----------------------------------------------------------------
      // Give time statistics
      //-----------------------------------------------------------------
      GenericAPI::queueLogEntry( msg.str( ),
                                 GenericAPI::LogHTML::MT_DEBUG,
                                 20,
                                 caller,
                                 "SCAN_MOUNTPT" );
      return error_str;
   }
   

   //--------------------------------------------------------------------
   /// This method perminately removes a mount point from the hash
   /// of mount points.
   ///
   /// \todo Need to investigate the efficientcy of doing the cleanup.
   //--------------------------------------------------------------------
   const std::list< std::string > MountPointHash::
   removeEntry( const std::string& dir, LogMessage& errors )
   {
      // Remove mount point from frame hash
      {
         ReadWriteLock lock( mLock, ReadWriteLock::WRITE,
                             __FILE__, __LINE__ );   
         mMountPtHash.erase( dir );    
      }   
   
   
      DirectoryHash::dir_container_type rm_dirs;

      DirectoryHash::Instance().removeDirectory( dir, rm_dirs, errors );
      return rm_dirs;
   }

   
   //--------------------------------------------------------------------
   /// This method removes all information from the mount point hash.
   //--------------------------------------------------------------------
   void MountPointHash::
   removeEntries()
   {
      // Destroy existing data hash in memory
      ReadWriteLock lock( mLock, ReadWriteLock::WRITE,
                          __FILE__, __LINE__ );   

      cleanup(); 
  
   
      return;
   }   
   
   //--------------------------------------------------------------------
   /// This method gets all known "IFO-Type-" frame data for a particular
   /// MOUNT_PT entry.
   ///   
   // FIX PR 2390: new query method: pass start,stop into search
   // Created overloaded functions; version 2 of 2, 
   // with query_start and query_stop as params
   //--------------------------------------------------------------------
   const std::vector< TimeHash::KnownIntervals > MountPointHash::
   getMountPtHashes( const std::string& ifo_type,
                     const std::string& mount_pt_dir,
                     const INT_4U query_start, const INT_4U query_stop,
                     const dir_info_type except_dir )
   {
      // Result data
      std::vector< TimeHash::KnownIntervals > known_data;
   
   
#if WORKING
      // Use mount_pt_dir key to get the IFOTypeHash value from the mMountPtHash
      MountPtConstIter mount_iter( mMountPtHash.find( mount_pt_dir ) );

      if( mount_iter == mMountPtHash.end() )
      {
         // mount_pt_dir doesn't have any data registered
         return known_data;
      }
   
   
      // Use the ifo_type key to get the vector of DirectoryInfo*'s value
      // from the IFOTypeHash
      IFOTypeConstIter
         ifo_iter( ( mount_iter->second ).find( ifo_type ) );

      if( ifo_iter == ( mount_iter->second ).end() )
      {
         // mount_pt_dir doesn't have any frame of "ifo-type" pair registered
         return known_data;
      }

      const vector< const DirectoryInfo* >& known_dirs( ifo_iter->second );   

      // Loop over vector of directories
      for( vector< const DirectoryInfo* >::const_iterator iter=known_dirs.begin(),
              end_iter = known_dirs.end(); iter != end_iter; ++iter )
      {
         if( *iter != except_dir )
         {
            // FIX PR 2390: new query method: pass start,stop into search
            known_data.push_back( ( *iter )->getFrameQueryIntervals( 
                                                                    ifo_type, query_start, query_stop ) );
         }
      }   
#endif /* WORKING */
   
   
      return known_data;
   }
   // FIX PR 2390: new query method: pass start,stop into search


#if old
   //------------------------------------------------------------------------------
   //
   //: Get ASCII representation of frame data with known ifo (could be "all") and 
   //: type (could be "all") from all MOUNT_PT directories.
   //      
   // This method gets ASCII representation of frame data with known "ifo" and
   // "type" from all MOUNT_PT entries in Tcl-layer requested format
   // (returned information will preserve the order in which directories appear
   // under MOUNT_PT list: rsc file variable).
   //   
   //!param: const std::string& ifo - An IFO for frame data to look up.
   //!param: const std::string& type - A Type for frame data to look up.
   //   
   //!return: const std::string - A list of all identified subdirectories with
   //+        time ranges for "ifo" and "type" frame data in a Tcl-layer requested
   //+        format.   
   //   
   const std::string MountPointHash::
   getTclHash( const std::string& ifo,
               const std::string& type )
   {
      string ascii;
   
#if WORKING   
      IfoTypeMatch match( ifo, type );
   
      ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                          __FILE__, __LINE__ );      
   
   
      // Step through all MOUNT_PT entries, look up searched data
      for( vector< string >::const_iterator
              mount_iter = mMountPtList.begin(),
              mount_end_iter = mMountPtList.end(); 
           mount_iter != mount_end_iter;
           ++mount_iter )
      {
         // Does mount point have any frame data registered?
         MountPtConstIter hash_iter( mMountPtHash.find( *mount_iter ) );
   
         if( hash_iter != mMountPtHash.end() && 
             hash_iter->second.empty() == false )
         {
            // Step through all registered frame data of "IFO-Type-":
            const IFOTypeHash& data( hash_iter->second );
   
   
            for( IFOTypeConstIter data_iter = data.begin(), data_end_iter = data.end();
                 data_iter != data_end_iter; ++data_iter )
            {
               // Examine "IFO-Type-" key of the map: 
               // if it matches searched criteria
               // then get all known data from all registered directories 
               // for this "IFO-Type-"
               if( match( data_iter->first ) )
               {
                  const vector< const DirectoryInfo* >& directories( data_iter->second );
   
                  for( vector< const DirectoryInfo* >::const_iterator dir_iter = directories.begin(),
                          dir_end_iter = directories.end();
                       dir_iter != dir_end_iter; ++dir_iter )
                  {
                     ascii += ( *dir_iter )->getTclHash( data_iter->first );
                  }
               } // matched searched ifo and type
            } // step through all registered "IFO-Type-" keys of the mount point entry
         } // mount point entry has frame data
      }

#endif /* WORKING */   
      return ascii;
   }
#endif /* old */

   //------------------------------------------------------------------------------
   //
   //: Get ASCII representation for mount point numbers with known ifo 
   //: (could be "all") and type (could be "all") from all MOUNT_PT directories.
   //      
   // This method gets ASCII representation of frame data with known "ifo" and
   // "type" from all MOUNT_PT entries in Tcl-layer requested format:
   //  { mount_point_name number_of_directories number_of_files} ... { } 
   //
   // (returned information will preserve the order in which directories appear
   // under MOUNT_PT list: rsc file variable).
   //   
   //!param: const std::string& ifo - An IFO for frame data to look up.
   //!param: const std::string& type - A Type for frame data to look up.
   //   
   //!return: const std::string - A list of all identified subdirectories with
   //+        "ifo" and "type" frame data in a Tcl-layer requested format:   
   //+ { mount_point_name number_of_directories number_of_files} ... { }
   //   
   const std::string MountPointHash::
   getHashNumbers( const std::string& ifo,
                   const std::string& type )
   {
      Query                     sp;
      // std::ostringstream        sc;

      // sc << ifo << "-" << type << "-";
      // sp.IFOType( sc.str( ) );
      sp.IFO( ifo );
      sp.Type( type );

      string ascii;

      bool complete = false;

      do {
         try
         {
            //--------------------------------------------------------------
            // Have to be very careful to avoid deadlock. Since this routine
            // is responsible for the ascii representation, will consider
            // not obtaining a lock less critical than other routines.
            //
            // In the case where a lock cannot be obtained, purge all 
            // results and start over.
            //--------------------------------------------------------------
            ascii = "";
            ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                                __FILE__, __LINE__ );
   
            QueryResultsMountPointStats       results( QueryResults::FORMAT_TCL );

            Find( sp, results );

            results.Finalize( );
            results.Swap( ascii );

            complete = true;
         }
         catch( const ReadWriteLock::BusyError& Error )
         {
            // Will loop back around and try again.
         }
      } while( ! complete );

      return ascii;
   }

   //------------------------------------------------------------------------------
   //
   //: Get ASCII representation of frame data intervals with known ifo 
   //: (could be "all") and type (could be "all") and within the bounds of 
   //: query_start and query_stop from all MOUNT_PT directories.
   //      
   //!param: const std::string& ifo - An IFO for frame data to look up.
   //!param: const std::string& type - A Type for frame data to look up.
   //!param: const unsigned int query_start - query start time.
   //!param: const unsigned int query_stop - query stop time.
   //   
   //!return: const std::string - A list of all identified subdirectories with
   //+        time ranges for "ifo" and "type" frame data in a Tcl-layer requested
   //+        format:
   //+  IFO-Type1 { i1_start i1_stop ... iN_start iN_stop } ... IFO_TypeN { ... }
   //   
   const std::string MountPointHash::
   getIntervalsList( const std::string& ifo, 
                     const std::string& type, 
                     const INT_4U query_start, 
                     const INT_4U query_stop )
   {
      ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                          __FILE__, __LINE__ );

      string return_string("");

      Query                     sp;

      sp.IFO( ifo );
      sp.Type( type );
      sp.StartTime( query_start );
      sp.EndTime( query_stop );
      sp.Extension( ".gwf" );   /// \todo remove hard coding of .gwf

      QueryResultsIntervals     results( QueryResults::FORMAT_TCL);

      Find( sp, results );

      results.Finalize( );
      results.Swap( return_string );

      return return_string;
   }


   //------------------------------------------------------------------------
   /// Get ASCII representation of frame data file names with known ifo 
   /// (could be "all") and type (could be "all") and within the bounds of 
   /// query_start and query_stop from all MOUNT_PT directories.
   //------------------------------------------------------------------------
   const std::string MountPointHash::
   getFileNames( const char* ifo_type_str,
                 const INT_4U query_start, 
                 const INT_4U query_stop )
   {
      ReadWriteLock     lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
                              __FILE__, __LINE__ );

      string            return_string("");
      Query             sp;

      sp.IFOTypeList( ifo_type_str );
      sp.StartTime( query_start );
      sp.EndTime( query_stop );
      sp.Extension( ".gwf" );   /// \todo remove hard coding of .gwf

      QueryResultsFilenames     results( QueryResults::FORMAT_TCL );

      Find( sp, results );

      results.Finalize( );
      results.Swap( return_string );

      return return_string;
   }


   
   //------------------------------------------------------------------------------
   //
   //: Get list of file extensions (as they appear in rsc file variable).
   //
   //!return: const std::vector< std::string > - A list of known file extensions.
   //
   const std::list< std::string > MountPointHash::getFileExtList()
   {
      ReadWriteLock lock( mLockFileExtList,
                          ReadWriteLock::READ, RWLOCK_TIMEOUT,
                          __FILE__, __LINE__ );

      // Just in case ... 
      if (mFileExtList.empty() == true) 
      {
         string exc( "MountPointHash::getFileExtList():" );
         exc += "MountPointHash class data member (mFileExtList) is empty.";
         exc += "Call updateFileExtList() to set list.";
         throw runtime_error( exc );
      }

      return mFileExtList;
   }


   
   //------------------------------------------------------------------------------
   //
   //: Create excluded directories message header.
   //      
   // This method creates a header for excluded directories message based on passed
   // to it directory pattern.   
   //   
   //!param: const std::list< std::string >& dir_pattern - A reference to the
   //+       pattern used to exclude directories.   
   //   
   //!return: const std::string - A generated message header.
   //   
   const std::string MountPointHash::createExcludedMsgHeader(
                                                                        const std::list< std::string >& dir_pattern )
   {   
      static const string header( "Directories excluded by pattern \'" ); 
   
      ostringstream pattern_stream;
      pattern_stream << header;
   
      copy( dir_pattern.begin(), dir_pattern.end(),
            ostream_iterator< string >( pattern_stream, " " ) );   
   
      pattern_stream << "\': ";   
   
   
      return pattern_stream.str();
   }
   
   
   //------------------------------------------------------------------------------
   //
   //: Write MOUNT_PT hash to the cache file.
   //      
   // This method writes a list of known MOUNT_PT entries to the cache file, and
   // initiates data hash writes of all known directories (recursively).   
   //   
   //!param: CacheFile* stream - A pointer to the cache file object to write to.
   //   
   //!return: Nothing.
   //   
   void MountPointHash::
   write( CacheFile& stream )
   {
      const mount_point_container_type
         mpts( MountPoints( ).Var( ) );

      stream << mpts
             << ExcludedDirs( ).Var( )
         ;
      
      // Write each MOUNT_PT entry to the stream recursively
      //   (with all identified subdirectories)
      for( mount_point_container_type::const_iterator
              iter = mpts.begin(), 
              end_iter = mpts.end();
           iter != end_iter;
           ++iter )
      {
         //--------------------------------------------------------------
         // Be cooperative with cancelation process
         //--------------------------------------------------------------
         Thread::Self::CancellationCheck( "diskCache::MountPointHash::write",
                                             __FILE__, __LINE__ );
         std::list< std::string > dirs( 1, *iter );
         DirectoryHash::Instance().WriteDirectories( dirs, stream );      
      }
   
      return;
   }

   
   //------------------------------------------------------------------------------
   //
   //: Read MOUNT_PT hash from the cache file.
   // 
   // This method reads a list of known MOUNT_PT entries from the cache file, and
   // initiates frame hash read for each directory listed in the same file.   
   //   
   // When calling this method, Tcl should guarantee that there're no threads in
   // the system that require an access to the global data hash (getFrameFiles_t,
   // getDirEnt_t, updateMountPtList, getDirCache).   
   //   
   //!param: CacheFile* stream - A pointer to the cache file object to write to.
   //   
   //!return: Nothing.
   //   
   void MountPointHash::read( CacheFile& stream )
   {
      static const char caller[] = "GenericAPI::MountPointHash::read";
      try
      {
         //--------------------------------------------------------------
         // Need to suspend the scanning of mount points.
         //--------------------------------------------------------------
         std::ostringstream     msg;

         msg << "Setting up for read"
            ;
         GenericAPI::queueLogEntry( msg.str( ),
                                    GenericAPI::LogHTML::MT_DEBUG,
                                    20,
                                    caller,
                                    "SCAN_MOUNTPT" );
         MountPointScanner::SyncRequest sync_request;
         MountPointScanner::scanner_sync_rw_type
            scanner_sync_lock( MountPointScanner::Sync( ) );

         // Invoke recursive cleanup of all objects that represent already
         // built hashes in memory (if any).
         {
            cleanup( );
     
   
            stream
               // Read a list of known MOUNT_PT entries
               >> mount_points( ).Var( )
               // Read a list of excluded directory names
               >> excluded_dirs( ).Var( );

            MountPointStatus::Set( MountPoints( ).Var( ) );
   
            // Read all directory entries from the stream
            DirectoryHash::Instance().readDirectories( stream );      
         }
      }
      catch( const std::exception& Exception )
      {
         std::ostringstream     msg;

         msg << "Unable to read cache file: " << Exception.what( );
         
         GenericAPI::queueLogEntry( msg.str( ),
                                    GenericAPI::LogHTML::MT_ORANGE,
                                    0,
                                    caller,
                                    "CACHE_READ" );
      }
   
   
      return;
   }
   
   
   //------------------------------------------------------------------------------
   //
   //: Validate list of MOUNT_PT entries.
   //    
   // This method guarantees that neither MOUNT_PT entry is a subdirectory of 
   // another MOUNT_PT entry.
   // It reports identified subdirectories in conflict as an error to the Tcl-layer,
   // and excludes that entry from mMountPtList data. It's Tcl's responsibility to
   // exclud reported subdirectory from it's own copy of the list (so deep scan
   // of that subdirectory is not issued).   
   //
   //!param: std::vector< std::string >& mount_pt_list - A reference to the 
   //+       new list of MOUNT_PT entries to validate.
   //   
   //!return: const std::string - Tcl formatted error message with identified 
   //+        conflicts.
   //
   // static   
   const std::string MountPointHash::
   validateMountPtList( std::vector< std::string >& mount_pt_list )
   {
      // Accumulative error message
      std::string errors;
   
#if PURGE      
      // Use plain while(...) instead of erase( remove_if(...) ) because of
      // iterator invalidation headache
      vector< string >::iterator iter( mount_pt_list.begin() );
   
      while( iter != mount_pt_list.end() )
      {
         // Append '/' to guarantee exact parent directory match, and to
         // avoid comparing to itself
         const string parent( *iter );
   
   
         vector< string >::iterator remove_iter( remove_if( mount_pt_list.begin(),
                                                            mount_pt_list.end(),
                                                            LogParent( parent, errors ) ) );

         if( remove_iter != mount_pt_list.end() )
         {
            // Subdirectories for "parent" were identified
            mount_pt_list.erase( remove_iter, mount_pt_list.end() );
   
            // "iter" might have been invalidated, find entry for "parent",
            // and proceed to the next MOUNT_PT
            iter = find( mount_pt_list.begin(), mount_pt_list.end(), parent );
         }

   
         ++iter;
      }
#else
      //-------------------------------------------------------------------------
      // Loop over the mount point list twice.
      // 1) Seed the validator with key pieces of information.
      // 2) Remove entries that do not conform to the standard.
      //-------------------------------------------------------------------------
      validator v( mount_pt_list );

      std::vector< std::string >::iterator
         new_end = remove_if( mount_pt_list.begin( ),
                              mount_pt_list.end( ),
                              v );
      mount_pt_list.erase( new_end, mount_pt_list.end( ) );
      v.Errors( errors );

#endif /* PURGE */
   
   
      return errors;
   }
   

   //------------------------------------------------------------------------------
   //
   //: Tcl-friendly formatted result data.
   //    
   //!param: std::list< std::string >& data - Result data.
   //!param: const std::string& errors - Error message.
   //!param: const std::string& header - Header for the result message. Default is 
   //+       an empty string.
   //
   //!return: const std::string - Tcl formatted string.
   //
   // static   
   const std::string MountPointHash::
   formatOutput( std::list< std::string >& data,
                 const std::string& errors, 
                 const std::string& header )
   {
      // Tcl expects sorted list
      data.sort();

   
      // Don't use ostream_iterator to format result string for Tcl:
      // Tcl doesn't like spaces between ' {' and entry, entry and '} '.

      string result;

      // FIX for PR 2603: opening brace needs preceding space
      result += " {"; 

      if( data.empty() == false )
      {
         result += header;
   
         list< string >::const_iterator iter( data.begin() );
   
         result += ( *iter );
         ++iter;

   
         for( list< string >::const_iterator end_iter = data.end();
              iter != end_iter; ++iter )
         {
            result += ' ';
            result += ( *iter );
         }
      }
   
      result += "} {";
      result += errors;
      result += "} ";
   
      return result;
   }

   
   //------------------------------------------------------------------------------
   //
   //: Object cleanup.
   // 
   // Caller must lock the object before calling this method!
   //   
   //!return: Nothing.
   //
   void MountPointHash::cleanup()
   {
      // Cleanup current MOUNT_PT list
      mount_points( ).Var( ).clear();   
   
      // Cleanup current list of excluded directory names
      mExcludedDirList.clear();
  
      // Cleanup hash of known data
      mMountPtHash.clear();
   
      return;
   }

   
} // namespace - diskCache

static bool
initialize( )
{
   General::AtExit::Append( diskCache::MountPointHash::removeEntries,
                            "diskCache::MountPointHash::removeEntries",
                            75 );
   return true;
}

validator::
validator( const std::vector< string >& Source )
   : m_original_directories( Source.begin( ), Source.end( ) ),
     m_parent_directories( Source.begin( ), Source.end( ) ),
     m_state( new state_type )
{
   //------------------------------------------------------------
   // Assume all elements are parent directories
   //------------------------------------------------------------
   m_parent_directories.sort( );
   //------------------------------------------------------------
   // Since m_parent_directories is ordered, it is now
   // easy to check if the neighbor is a child
   //------------------------------------------------------------
   if ( m_parent_directories.size( ) > 2 )
   {
      std::list< std::string >::const_iterator
         cur = m_parent_directories.begin( ),
         pre = m_parent_directories.begin( ),
         last = m_parent_directories.end( );
      ++cur;
      while( cur != last )
      {
         this->operator( )( *cur, *pre );
         ++cur;
         ++pre;
      }
   }
   //------------------------------------------------------------
   // Advance the state since initialization has occurred.
   //------------------------------------------------------------
}

void validator::
Errors( std::string& Message ) const
{
   if ( ( m_state->m_child_directories.size( ) > 0 )
        || ( m_state->m_duplicate_directories.size( ) > 0 ) )
   {
      Message = "{ For the mount point list: ";
      {
         print_element  pe( Message, ',' );

         std::for_each( m_original_directories.begin( ),
                        m_original_directories.end( ),
                        pe );
      }
      if ( m_state->m_child_directories.size( ) > 0 )
      {
         print_element  pe( Message, ',' );

         Message += " Removed the following sub directories: ";

         std::for_each( m_state->m_child_directories.begin( ),
                        m_state->m_child_directories.end( ),
                        pe );
      }
      if ( m_state->m_duplicate_directories.size( ) > 0 )
      {
         print_element  pe( Message, ',' );

         Message += " Removed the following duplicate directories: ";

         std::for_each( m_state->m_duplicate_directories.begin( ),
                        m_state->m_duplicate_directories.end( ),
                        pe );
      }
      Message += " }";
   }
}

bool validator::
operator()( const element_type& Current,
            const element_type& Previous )
{
   bool retval = true;

   if ( is_parent( Current, Previous ) )
   {
      //---------------------------------------------------------
      // Construct a list of child directories.
      //---------------------------------------------------------
      m_state->m_child_directories.insert( Current );
      retval = false;
   }

   return retval;
}

bool validator::
operator()( const std::string& MountPoint )
{
   //------------------------------------------------------------
   // 1) Check to see if the directory has been flagged as a child
   // 2) Check the occurrances of the directory within the list.
   //------------------------------------------------------------
   const INT_4U c = ++( m_state->m_dir_count[ MountPoint ] );

   if ( c == 2 )
   {
      m_state->m_duplicate_directories.insert( MountPoint );
   }

   return ( ( m_state->m_child_directories.find( MountPoint )
              != m_state->m_child_directories.end( ) )
            || ( c > 1 ) );
}

inline bool validator::
is_parent( const element_type& Current,
           const element_type& Previous )
{
   const size_t       pl = Previous.length( );

   //------------------------------------------------------------
   // For one directory to be a sub directory of another,
   // The following conditions need to apply:
   // 1) The child directory needs to be longer than the parent.
   // 2) The child directory needs to have a separator
   //    character at the position where the parent ends.
   // 3) The parent directory must be a substring of the
   //    child.
   //------------------------------------------------------------
   return ( ( Current.length( ) > pl )
            && ( Current[ pl ] == '/' )
            && ( Current.compare( 0, pl, Previous ) == 0 ) );
}
