#ifdef HAVE_CONFIG_H
#include "genericAPI/config.h"
#endif /* HAVE_CONFIG_H */

#include <signal.h>

#include <sstream>

#include "general/ErrorLog.hh"
#include "general/TaskThread.hh"
#include "general/Timeout.hh"

#include "genericAPI/MountPointStatus.hh"
#include "genericAPI/StatVFS.hh"
#include "genericAPI/StatVFSDirect.hh"
#include "genericAPI/StatVFSPool.hh"


using General::StdErrLog;
using General::Timeout;
using GenericAPI::MountPointStatus;
using GenericAPI::StatVFSPool;

namespace GenericAPI
{
  extern int lstat_interupt( );
  extern int lstat_timeout( );
}

namespace
{
  class timed_statvfs
    : public General::Task
  {
  public:
    timed_statvfs( const std::string& Path, struct statvfs& StatVFSBuf )
      : Task( GenericAPI::lstat_interupt( ) ),
	m_path( Path ),
	m_statvfsbuf( StatVFSBuf ),
	m_err( -1 ),
	m_errno( EINTR )
	
    {
    }

    inline int Error( ) const
    {
      return m_err;
    }
    
    inline int ErrNo( ) const
    {
      return m_errno;
    }

    virtual void operator()( )
    {
      if ( StdErrLog.IsOpen( ) )
      {
	std::ostringstream	msg;
      
	msg << "timed_statvfs: thread: " << (void*)pthread_self( )
	    << " lstat(" << m_path << ")";
	StdErrLog( General::ErrorLog::DEBUG,
		   __FILE__, __LINE__,
		   msg.str( ) );
	struct timespec wakeup;

	wakeup.tv_sec = 1;
	wakeup.tv_nsec = 0;
	m_err = ( nanosleep( &wakeup, NULL ) != 0 );
      }

      m_statvfs = StatVFSPool::Request( );
      try
      {
	m_err = m_statvfs->StatVFS( m_path, m_statvfsbuf );
      }
      catch( const std::exception& Exception )
      {
	m_err = -1;
	if ( StdErrLog.IsOpen( ) )
	{
	  StdErrLog( General::ErrorLog::DEBUG,
		     __FILE__, __LINE__,
		     Exception.what( ) );
	}
      }
      catch( ... )
      {
	// :TODO: Unknown exception
	m_err = -1;
      }
      m_errno = errno;
      StatVFSPool::Release( m_statvfs );
      m_statvfs = (StatVFSPool::info_type)NULL;
      if ( StdErrLog.IsOpen( ) )
      {
	std::ostringstream	msg;
      
	msg << "timed_statvfs: thread: " << (void*)pthread_self( )
	    << " m_err: " << m_err
	    << " m_errno: " << m_errno;
	StdErrLog( General::ErrorLog::DEBUG,
		   __FILE__, __LINE__,
		   msg.str( ) );
      }
    }

  private:
    const std::string		m_path;
    struct statvfs&		m_statvfsbuf;
    int				m_err;
    int				m_errno;
    StatVFSPool::info_type	m_statvfs;

    static void
    destroyer( timed_statvfs* TStatVFS )
    {
      volatile StatVFSPool::info_type rstatvfs = TStatVFS->m_statvfs;

      TStatVFS->m_statvfs = (StatVFSPool::info_type)NULL;
      //-----------------------------------------------------------------
      // This is called when things are taking too long
      //-------------------------------------------------------------------
      if ( StdErrLog.IsOpen( ) )
      {
	std::ostringstream	msg;
	
	msg << "destroyer: Stat buffer: " << (void*)rstatvfs;
	StdErrLog( General::ErrorLog::DEBUG,
		   __FILE__, __LINE__,
		   msg.str( ) );
	struct timespec wakeup;
      
	wakeup.tv_sec = 1;
	wakeup.tv_nsec = 0;
	nanosleep( &wakeup, NULL );
      }
      StatVFSPool::Destroy( rstatvfs );
      TStatVFS->m_err = -1;
      TStatVFS->m_errno = EINTR;
    }
  };
} // namespace - anonymous

namespace GenericAPI
{
  int
  StatVFS( const std::string& Filename, struct statvfs& StatVFSBuffer )
  {
    int retval = -1;

    if ( MountPointStatus::Status( Filename )
	 == MountPointStatus::STATE_OFFLINE )
    {
      errno = EINTR;
      return retval;
    }

    //-------------------------------------------------------------------
    // PR#3341 -
    // Be careful with the statvfs call since it can lock up under
    //   nfs usage and not return until the server is once again
    //   available.
    // This condition is avoided by setting a maximum time for this
    //   call to return before causing the statvfs call to be interupted.

    timed_statvfs*	tstatvfs
      = new timed_statvfs( Filename, StatVFSBuffer );

    try
    {
      Timeout( tstatvfs, lstat_timeout( ) );

      retval = tstatvfs->Error( );
      errno = tstatvfs->ErrNo( );
      delete tstatvfs;
    }
    catch( ... )
    {
      errno = EINTR;
    }

    if ( errno == EINTR )
    {
      MountPointStatus::Offline( Filename );
    }

    return retval;
  }
}

